"""
CLI entry point
"""
import logging
from decimal import Decimal
import click
import click_log

from pricedb import PriceDbApplication
# from pricedb.map_cli import symbol_map
from pricedb.model import PriceModel, SecuritySymbol
from pricedb.config_cli import config_cli

logger = logging.getLogger(__name__)
click_log.basic_config(logger)


@click.group()
@click_log.simple_verbosity_option(logger)
def cli():
    """ entry point """
    pass

@click.command()
@click_log.simple_verbosity_option(logger)
def config():
    ''' Display the configuration values '''
    app = PriceDbApplication()
    #app.

@click.command()
@click.option("--symbol", prompt="Security symbol with exchange (EXCH:SYMBOL)", type=str)
@click.option("--date", prompt="Date in ISO format (yyyy-mm-dd)")
@click.option("--value", prompt="Price")
@click.option("--currency", prompt=True, type=str)
@click_log.simple_verbosity_option(logger)
def add(symbol: str, date, value, currency: str):
    """ Add individual price """
    symbol = symbol.upper()
    currency = currency.upper()

    app = PriceDbApplication()
    price = PriceModel()

    # security = SecuritySymbol("", "")
    price.symbol.parse(symbol)
    # price.symbol.mnemonic = price.symbol.mnemonic.upper()

    # date_str = f"{date}"
    # date_format = "%Y-%m-%d"
    # if time:
    #     date_str = f"{date_str}T{time}"
    #     date_format += "T%H:%M:%S"
    # datum.from_iso_date_string(date)
    # price.datetime = datetime.strptime(date_str, date_format)
    price.datum.from_iso_date_string(date)

    price.value = Decimal(value)
    price.currency = currency
    app.add_price(price)
    app.save()

    click.echo("Price added.")


@click.command()
#@click.option("--format", help="The export format. The default is ledger.", default="ledger")
@click_log.simple_verbosity_option(logger)
def export():
    """ Export prices in ledger format """
    app = PriceDbApplication()
    result = app.ledger_export()
    print(result)


@click.command("import")
@click.argument("filepath") # "FILE - path to the .csv file to import"
@click.argument("currency") # "currency - to use for imported prices"
@click_log.simple_verbosity_option(logger)
def import_csv(filepath: str, currency: str):
    """ Import prices from CSV file """
    logger.debug(f"currency = {currency}")
    # auto-convert to uppercase.
    currency = currency.upper()

    app = PriceDbApplication()
    app.logger = logger
    app.import_prices(filepath, currency)


@click.command()
@click.option("symbol", "-s")
@click_log.simple_verbosity_option(logger)
def last(symbol: str):
    """ displays last price, for symbol if provided """
    app = PriceDbApplication()

    # convert to uppercase
    if symbol:
        symbol = symbol.upper()
        # extract namespace
        sec_symbol = SecuritySymbol("", "")
        sec_symbol.parse(symbol)

        latest = app.get_latest_price(sec_symbol)
        assert isinstance(latest, PriceModel)
        print(f"{latest}")
    else:
        # Show the latest prices available for all securities.
        latest = app.get_latest_prices()
        for price in latest:
            print(f"{price}")


@click.command("list")
@click.option("--date", help="The date for which to show prices.")
@click.option("--currency", help="The currency for which to show prices.")
@click.option("--last", is_flag=True, help="Display only the latest prices per symbol")
@click_log.simple_verbosity_option(logger)
def list_prices(date, currency, last):
    """ Display all prices """
    app = PriceDbApplication()
    app.logger = logger

    if last:
        # fetch only the last prices
        prices = app.get_latest_prices()
    else:
        prices = app.get_prices(date, currency)
    for price in prices:
        print(price)

    print(f"{len(prices)} records found.")


@click.command("dl")
@click.option("--exchange", "-x", default=None, help="Exchange for the securities to update")
@click.option("--symbol", "-s", default=None, help="Symbol for the individual price to download")
@click.option("--agent", "-a", default=None,
              help="Agent to use for download (vanguard_au, morningstar, alphavantage)")
@click.option("--currency", "-c", default=None, help="Currency symbol to use for the price(s)")
@click.option("--help", "-h", is_flag=True)
@click_log.simple_verbosity_option(logger)
@click.pass_context
def download(ctx, help: bool, symbol: str, exchange: str, agent: str, currency: str):
    """ Download the latest prices """
    if help:
        click.echo(ctx.get_help())
        ctx.exit()

    app = PriceDbApplication()
    app.logger = logger

    if currency:
        currency = currency.strip()
        currency = currency.upper()

    # Otherwise download the prices for securities listed in the database.
    app.download_prices(currency=currency, agent=agent, symbol=symbol, exchange=exchange)


@click.command("prune")
@click.option("--all", "-a", is_flag=True, help="Delete historical prices for all symbols")
@click.option("--symbol", "-s", help="Symbol for which to delete prices")
@click_log.simple_verbosity_option(logger)
def prune(symbol: str, all: str):
    """ Delete old prices, leaving just the last. """
    app = PriceDbApplication()
    app.logger = logger
    count = 0

    if symbol is not None:
        sec_symbol = SecuritySymbol("", "")
        sec_symbol.parse(symbol)

        deleted = app.prune(sec_symbol)
        if deleted:
            count = 1
    else:
        count = app.prune_all()

    print(f"Removed {count} old price entries.")


@click.command("securities")
@click_log.simple_verbosity_option(logger)
def securities():
    ''' export the list of securities for which the prices are tracked '''
    app = PriceDbApplication()
    result = app.get_security_list()
    print(result)


######
cli.add_command(add)
cli.add_command(export)
cli.add_command(download)
cli.add_command(import_csv)
# cli.add_command(symbol_map)
cli.add_command(last)
cli.add_command(list_prices)
cli.add_command(prune)
cli.add_command(securities)

cli.add_command(config_cli)



""" main api point """
import logging
from typing import List

from . import dal, mappers
from .model import PriceModel, SecuritySymbol


class PriceDbApplication:
    """ Contains the main public interfaces """

    def __init__(self, session=None):
        self.logger = logging.getLogger(__name__)
        self.price_repo = None
        self.security_repo = None
        self.__session = session

    def add_price(self, price: PriceModel):
        """ Creates a new price record """
        # assert isinstance(price, PriceModel)

        if not price:
            raise ValueError("Cannot add price. The received model is null!")

        mapper = mappers.PriceMapper()
        entity = mapper.map_model(price)

        self.add_price_entity(entity)

    def add_price_entity(self, price: dal.Price):
        """ Adds the price """
        from decimal import Decimal

        # check if the price already exists in db.
        repo = self.get_price_repository()
        existing = (
            repo.query
            .filter(dal.Price.namespace == price.namespace)
            .filter(dal.Price.symbol == price.symbol)
            .filter(dal.Price.date == price.date)
            .filter(dal.Price.time == price.time)
            .first()
        )
        if existing:
            # Update existing price.
            new_value = Decimal(price.value) / Decimal(price.denom)
            self.logger.info(f"Exists: {price}")
            if price.currency != existing.currency:
                raise ValueError(
                    f"The currency is different for price {price}!")
            if existing.value != price.value:
                existing.value = price.value
                self.logger.info(f"Updating to {new_value}.")
            if existing.denom != price.denom:
                existing.denom = price.denom
        else:
            # Insert new price
            self.session.add(price)
            self.logger.info(f"Added {price}")

    def download_price(self, symbol: str, exchange: str, currency: str, agent: str) -> PriceModel:
        """ Download and save price online """
        price = self.__download_price(symbol, exchange, currency, agent)
        self.save()
        return price

    def download_prices(self, **kwargs):
        """ Downloads all the prices that are listed in the Security table.
        Accepts filter arguments: currency, agent, symbol, exchange.
        """
        currency: str = kwargs.get('currency', None)
        if currency:
            currency = currency.upper()
        agent: str = kwargs.get('agent', None)
        if agent:
            agent = agent.upper()
        symbol: str = kwargs.get('symbol', None)
        if symbol:
            symbol = symbol.upper()
        exchange: str = kwargs.get('exchange', None)
        if exchange:
            exchange = exchange.upper()
        securities = self.__get_securities(currency, agent, symbol, exchange)
        #self.logger.debug(securities)

        for sec in securities:
            #symbol = f"{sec.namespace}:{sec.symbol}"
            symbol = sec.symbol
            currency = sec.currency
            agent = sec.updater
            exchange = sec.namespace
            #self.logger.debug(f"Initiating download for {symbol} {currency} with {agent}...")
            try:
                self.__download_price(symbol.strip(), exchange, currency, agent)
            except Exception as e:
                self.logger.error(str(e))
        self.save()

    def get_config_values(self):
        #Config
        pass

    def import_prices(self, file_path: str, currency_symbol: str):
        """ Incomplete """
        from .csv import CsvParser

        assert isinstance(file_path, str)
        assert isinstance(currency_symbol, str)

        self.logger.debug(f"Importing {file_path}")
        parser = CsvParser()
        prices = parser.parse_file(file_path, currency_symbol)

        counter = 0
        session = self.session
        # Create insert statements
        mapper = mappers.PriceMapper()
        for price in prices:
            new_price = mapper.map_model(price)
            self.add_price_entity(new_price)
            counter += 1
        # Save all to database
        session.commit()
        print(f"{counter} records inserted.")

    def get_latest_price(self, symbol: SecuritySymbol) -> PriceModel:
        """ Returns the latest price for the given symbol """
        from .repositories import PriceRepository

        # TODO should include the currency? Need a public model for exposing the result.
        assert isinstance(symbol, SecuritySymbol)

        session = self.session
        repo = PriceRepository(session)
        query = (
            repo.query
            .filter(dal.Price.symbol == symbol.mnemonic)
            .order_by(dal.Price.date.desc(), dal.Price.time.desc())
        )
        if symbol.namespace:
            query = query.filter(dal.Price.namespace == symbol.namespace)

        latest = query.first()

        # map
        mapper = mappers.PriceMapper()
        result = mapper.map_entity(latest)

        return result

    @property
    def session(self):
        """ Returns the current db session """
        if not self.__session:
            self.__session = dal.get_default_session()
        return self.__session

    def get_prices(self, date: str = None, currency: str = None) -> List[PriceModel]:
        """ Fetches all the prices for the given arguments from the database """
        from .repositories import PriceRepository

        session = self.session
        repo = PriceRepository(session)
        query = repo.query
        if date:
            query = query.filter(dal.Price.date == date)
        if currency:
            query = query.filter(dal.Price.currency == currency)
        # Sort by symbol.
        query = query.order_by(dal.Price.namespace, dal.Price.symbol)
        price_entities = query.all()

        mapper = mappers.PriceMapper()
        result = []
        for entity in price_entities:
            model = mapper.map_entity(entity)
            result.append(model)
        return result

    def get_prices_on(self, on_date: str, namespace: str, symbol: str):
        """ Returns the latest price on the date """
        repo = self.get_price_repository()
        query = (
            repo.query.filter(dal.Price.namespace == namespace)
            .filter(dal.Price.symbol == symbol)
            .filter(dal.Price.date == on_date)
            .order_by(dal.Price.time.desc())
        )
        result = query.first()
        # logging.debug(result)
        return result

    def get_latest_prices(self):
        """ Fetches the latest prices for all symbols """
        # get all symbols first, for which we have prices available
        from .repositories import PriceRepository

        repo = PriceRepository(self.__session)

        query = (
            repo.session.query(dal.Price.namespace, dal.Price.symbol)
            .order_by(dal.Price.namespace, dal.Price.symbol)
            .distinct()
        )
        all_symbols = query.all()
        # self.logger.debug(all_symbols)

        # Get the latest prices for these symbols
        latest_prices = []
        for symbol_price in all_symbols:
            symbol = SecuritySymbol(symbol_price.namespace, symbol_price.symbol)
            latest = self.get_latest_price(symbol)
            latest_prices.append(latest)
        return latest_prices

    def get_price_repository(self):
        """ Price repository """
        from .repositories import PriceRepository

        if not self.price_repo:
            self.price_repo = PriceRepository(self.session)
        return self.price_repo

    ## Securities

    def get_security_list(self) -> str:
        ''' retrieve the securities from the database '''
        from pricedb.repositories import SecurityRepository

        #repo = SecurityRepository(self.session)
        #all_symbols = repo.query.all()
        all_symbols = self.__get_securities(None, None, None, None)
        
        # sort by symbol
        all_symbols.sort(key=lambda sec: sec.symbol)

        output = ""
        for security in all_symbols:
            output += security.symbol + "\n"
        return output

    def ledger_export(self):
        ''' Export prices in ledger format '''
        from pricedb.ledger import LedgerFormatter
        from pricedb.config2 import Configuration

        # load all prices
        prices = self.get_prices()
        # sort by date
        prices.sort(key=lambda price: price.datum.datetime)

        # export in ledger format
        formatter = LedgerFormatter()
        result = formatter.format_prices(prices)

        # If the export destination is set, save there directly.
        cfg = Configuration()
        dest = cfg.export_destination
        if dest is not None:
            self.__save_text_file(dest, result)
            # Replace the output with the message instead.
            result = f'Data saved in {dest}'

        return result

    def prune_all(self) -> int:
        """
        Prune historical prices for all symbols, leaving only the latest.
        Returns the number of items removed.
        """
        from .repositories import PriceRepository

        # get all symbols that have prices
        repo = PriceRepository()
        items = repo.query.distinct(dal.Price.namespace, dal.Price.symbol).all()
        # self.logger.debug(items)
        count = 0

        for item in items:
            symbol = SecuritySymbol(item.namespace, item.symbol)
            deleted = self.prune(symbol)
            if deleted:
                count += 1

        return count

    def prune(self, symbol: SecuritySymbol):
        """
        Delete all but the latest available price for the given symbol.
        Returns the number of items removed.
        """
        from .repositories import PriceRepository

        assert isinstance(symbol, SecuritySymbol)

        self.logger.debug(f"pruning prices for {symbol}")

        repo = PriceRepository()
        query = (
            repo.query.filter(dal.Price.namespace == symbol.namespace)
            .filter(dal.Price.symbol == symbol.mnemonic)
            .order_by(dal.Price.date.desc())
            .order_by(dal.Price.time.desc())
        )
        all_prices = query.all()
        # self.logger.debug(f"fetched {all_prices}")

        deleted = False
        first = True
        for single in all_prices:
            if not first:
                repo.query.filter(dal.Price.id == single.id).delete()
                deleted = True
                self.logger.debug(f"deleting {single.id}")
            else:
                first = False

        repo.save()

        return deleted

    def save(self):
        """ Save changes """
        if self.__session:
            self.session.commit()
        else:
            self.logger.warning("Save called but no session open.")

    def __download_price(self, symbol: str, exchange: str, currency: str, agent: str):
        """ Downloads and parses the price """
        from finance_quote_python import Quote

        assert isinstance(symbol, str)
        assert isinstance(exchange, str)
        assert isinstance(currency, str)
        assert isinstance(agent, str)

        if not symbol:
            return None

        #self.logger.info(f"Downloading {symbol}... ")

        dl = Quote()
        dl.logger = self.logger

        dl.exchange = exchange
        dl.set_source(agent)
        dl.set_currency(currency)

        #result = dl.fetch(agent, [symbol])
        result = dl.fetch(exchange, [symbol])

        if not result:
            raise ValueError(f"Did not receive a response for {symbol}.")

        price = result[0]

        if not price:
            raise ValueError(f"Price not downloaded/parsed for {symbol}.")
        else:
            # Create price data entity, to be inserted.
            self.add_price(price)

        return price

    def __get_securities(self, currency: str, agent: str, symbol: str,
                         namespace: str) -> List[dal.Security]:
        '''
        Fetches the securities that match the given filters
        '''
        from .repositories import SecurityRepository
        #from .repositories import SecurityRepositoryTiny

        repo = SecurityRepository(self.session)
        #repo = SecurityRepositoryTiny()
        query = repo.query

        if currency is not None:
            query = query.filter(dal.Security.currency == currency)

        if agent is not None:
            query = query.filter(dal.Security.updater == agent)

        if symbol is not None:
            query = query.filter(dal.Security.symbol == symbol)

        if namespace is not None:
            query = query.filter(dal.Security.namespace == namespace)

        # Sorting
        query = query.order_by(dal.Security.namespace, dal.Security.symbol)

        securities = query.all()
        return securities

    def __save_text_file(self, path: str, content: str):
        ''' Saves a text file '''
        with open(path, 'w') as outfile:
            outfile.write(content)



'''
CLI for configuration manipulation
'''
import click
import click_log
import logging

logger = logging.getLogger(__name__)
click_log.basic_config(logger)


@click.group("cfg")
@click_log.simple_verbosity_option(logger)
def config_cli():
    """ Configuration management """
    pass


@click.command("show")
def show_config():
    ''' Display the current configuration '''
    from pricedb.config import package_name
    from usersconfig.configuration import Configuration

    reader = Configuration(package_name)
    cfg = reader.load()

    print(cfg)

####


config_cli.add_command(show_config)



""" Repositories - operations on multiple entities/aggregates """
# from typing import List
from .dal import Price, get_default_session, Security

DB_NAME = 'price-database.json'

class PriceRepository:
    """ Operations on prices """
    def __init__(self, session=None):
        self.__session = session

    def add(self, entity: Price):
        """ Insert price """
        self.session.add(entity)

    def save(self):
        """ Save all changes in the session """
        self.session.commit()

    @property
    def session(self):
        """ db session """
        if not self.__session:
            self.__session = get_default_session()
        return self.__session

    @property
    def query(self):
        """ Base query """
        return self.session.query(Price)


class SecurityRepository:
    """ Security table """
    def __init__(self, session=None):
        self.__session = session

    @property
    def session(self):
        """ db session """
        if not self.__session:
            self.__session = get_default_session()
        return self.__session

    @property
    def query(self):
        """ Base query """
        return self.session.query(Security)


class SecurityRepositoryTiny:
    """
    Securities.
    Implemented with tinydb. Uses the current folder for the database.
    Fields: id, symbol, namespace, updater, currency.
    """
    def __init__(self):
        from tinydb import TinyDB

        self.db_path = f'./{DB_NAME}'
        self.db = TinyDB(self.db_path)
        self.table = self.db.table('security')

    def all_securities(self):
        ''' Returns all records '''
        #from tinydb import Query
        #Sec = Query()
        result = self.table.all()
        return result



"""
Ledger-related code.
Used for export of prices in ledger format.
"""
from typing import List
from pricedb.model import PriceModel


class LedgerFormatter:
    ''' Format a price for ledger price db '''

    def format_price(self, price: PriceModel):
        '''
        ledger price format, ISO format supported:
        P 2004-06-21 02:17:58 TWCUX $27.76
        '''
        date_time = price.datum.to_datetime_string()

        fmt = f"P {date_time} {price.symbol.mnemonic} {price.value} {price.currency}"
        return fmt

    def format_prices(self, prices: List[PriceModel]):
        ''' format a list of prices '''
        output = ""
        for price in prices:
            output += self.format_price(price) + "\n"
        return output



'''
Store configuration in YAML format,
in user's configuration directory.
'''


class Configuration:
    def __init__(self):
        self.config_file_name = "config.yaml"
        #self.app_name = "pricedb"
        self.app_name = __name__
        self.create_config_file()
        self.config = self.read_config_content()

    def read_config_content(self):
        ''' Read configuration file '''
        import yaml

        path = self.getConfigPath()
        content = ''
        with open(path, 'r') as stream:
            try:
                content = yaml.safe_load(stream)
                # print(content)
            except yaml.YAMLError as exc:
                print(exc)

        return content

    def create_config_file(self):
        ''' create the config file if it does not exist '''
        from usersconfig.configuration import Configuration

        cfg = Configuration(__name__)
        cfg.create_default_config_file()

    # @property
    def get_config_dir(self):
        from xdg.BaseDirectory import xdg_config_home
        from os.path import sep

        return xdg_config_home + sep + self.app_name

    # @property
    def getConfigPath(self):
        ''' assembles the path to the config file '''
        from os.path import sep

        return self.get_config_dir() + sep + self.config_file_name

    @property
    def export_destination(self):
        value = self.config["export_destination"]
        return value



"""
Data layer
"""
from sqlalchemy import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy import Column, Integer, String

Base = declarative_base()


class Price(Base):
    """ Price entity """
    __tablename__ = 'price'

    id = Column(Integer, primary_key=True)
    namespace = Column(String)
    symbol = Column(String)
    date = Column(String)
    time = Column(String)
    value = Column(Integer)
    denom = Column(Integer)
    currency = Column(String)

    def __repr__(self):
        actual_value = 0
        if self.denom:
            actual_value = self.value / self.denom
        actual_value_str = f"{actual_value:.2f}"

        symbol = f"{self.namespace}:{self.symbol}" if self.namespace else self.symbol
        if symbol is None:
            symbol = ""
        formatted_symbol = f"{symbol:<13}"

        return f"<Price ({formatted_symbol},{self.date} {self.time},{actual_value_str:>6})>"


class Security(Base):
    """ The security / symbol entity
    Adding a record here should enable it for updated automatically.
    Contains the link to Yahoo symbol and should replace SymbolMap.
    """
    __tablename__ = "security"

    id = Column(Integer, primary_key=True)
    symbol = Column(String, unique=True)
    namespace = Column(String)
    updater = Column(String)
    currency = Column(String)

    def __repr__(self):
        return f"<Security (id={self.id},symbol={self.symbol})>"


def get_default_session():
    '''
    Return the default database session. The db file path is read from the config.
    Implementation using the Yaml config.
    '''
    from pricedb.config import package_name
    from usersconfig.configuration import Configuration
    from .config import ConfigKeys

    cfg = Configuration(package_name).load()
    # print(cfg)
    
    db_path = cfg[ConfigKeys.price_database_path.name]
    if not db_path:
        raise ValueError("Price database not set in the configuration file!")
    return get_session(db_path)

# def get_default_session():
#     """ Return the default session. The path is read from the default config. """
#     from .config import Config, ConfigKeys

#     db_path = Config().get(ConfigKeys.price_database_path)
#     if not db_path:
#         raise ValueError("Price database not set in the configuration file!")
#     return get_session(db_path)

def get_session(db_path: str):
    """ Creates and opens a database session """
    # connection
    con_str = "sqlite:///" + db_path
    # Display all SQLite info with echo.
    engine = create_engine(con_str, echo=False)

    # create metadata (?)
    Base.metadata.create_all(engine)

    # create session
    Session = sessionmaker(bind=engine)
    session = Session()

    return session
